unit ChainU;

{
  Dit programma bepaalt de exacte chain-lengte voor addition-chains voor een en
  twee elementen. Ook is het mogelijk om een exacte chain te laten creeren.

  We onderscheiden triviale en niet triviale chains. De triviale chains worden
  door het Brun-algorithme bepaald. De niet-triviale chains worden opgeslagen.
}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Spin, StdCtrls;

const
  MAXLLL = 10;
  MAXchain = 20;
  MAXL = 248;

type
  TLLL = array[1..MAXLLL] of longint;
  Tchain = array[0..MAXchain] of longint;
  TL = array[1..MAXL] of integer;
  TLL = array[1..MAXL,1..MAXL] of integer;

  TF_Chain = class(TForm)
    B_Calc: TButton;
    B_Exit: TButton;
    E_Chain: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    S_M: TSpinEdit;
    S_N: TSpinEdit;
    Label4: TLabel;
    E_Len: TEdit;
    B_Table: TButton;
    B_Debug: TButton;
    procedure B_CalcClick(Sender: TObject);
    procedure B_ExitClick(Sender: TObject);
    procedure B_TableClick(Sender: TObject);
    procedure S_Exit(Sender: TObject);
    procedure S_Change(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure B_DebugClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

procedure CalcLLTabel;
function calcLL(x, y, in_low, in_high : integer) : integer;
function calcLLLTable( var LLL : TLLL; len, in_low, in_exp, in_high : integer;
                       var out_low, out_high, Chain_Prec : integer) : boolean;

var
  F_Chain : TF_Chain;
  Chain : TChain;
  L : TL;
  LL, LLbrun, LLdiff, ChainData : TLL;
  maxLL : integer;
  
implementation

uses TableU, DebugU;

{$R *.DFM}

procedure showDebug( var LLL : TLLL; L, len : integer; bool : boolean;
                     txt : string);
var
  i : integer;
  s : string;
begin
  if (length(txt) > 1) OR (pos(txt, DebugU.S_Debug) = 0) then exit;
  s := txt + ': L(';
  for i := 1 to len - 1 do s := s + inttostr(LLL[i]) + ',';
  if bool
    then s := s + inttostr(LLL[len]) + ') = ' + inttostr(L)
    else s := s + inttostr(LLL[len]) + ') >= ' + inttostr(L);
  DebugU.F_Debug.M_Debug.Lines.Add(s);
  Application.ProcessMessages;
end;

function calcLLL( var LLL : TLLL; len, in_low, in_high : integer) : integer;
var
  LLL_low, LLL_high, new_len, i, j, x, L, Chain_Prec : integer;
  new_LLL : TLLL;
  done : boolean;
begin
  showDebug( LLL, in_low, len, TRUE, 'X');
  if len = 2 then
  begin
    calcLLL := calcLL(LLL[1], LLL[2], in_low, in_high);
    exit;
  end;
  if LLL[len] - LLL[len-1] < 3 then
  begin
    L := calcLLL(LLL, len - 1, in_low - 1, in_high - 1);
    inc(L);
    chain[L] := LLL[len];
    calcLLL := L;
    exit;
  end;

  j := len - 1;
  while (2 * LLL[j] >= LLL[len]) and ( j > 0) do
  begin
    i := 1;
    while LLL[i] + LLL[j] < LLL[len] do
      inc(i);
    if LLL[i] + LLL[j] = LLL[len] then
    begin
      L := calcLLL(LLL, len - 1, in_low - 1, in_high - 1);
      inc(L);
      chain[L] := LLL[len];
      calcLLL := L;
      exit;
    end;
    dec(j);
  end;
  LLL_high := -1;
  done := calcLLLTable( LLL, len, in_low, in_high, in_high,
                        LLL_low, LLL_high, Chain_Prec);
  showDebug( LLL, LLL_high, len, done, 'Y');
  if done then
  begin
    x := LLL[len] - Chain_Prec;
    i := 1;
    j := 1;
    while LLL[i] < x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if x > 2 then
    begin
      new_LLL[j] := x;
      inc(j);
    end;
    if LLL[i] = x then inc(i);
    while LLL[i] < LLL[len] - x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if 2 * x <> LLL[len] then
    begin
      new_LLL[j] := LLL[len] - x;
      inc(j);
    end;
    if LLL[i] = LLL[len] - x then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;

    L := calcLLL(new_LLL, new_len, in_low - 1, in_high - 1);
    showDebug( new_LLL, in_high - 1, new_len, TRUE, 'W');
    inc(L);
    chain[L] := LLL[len];
    calcLLL := L;;
  end
  else CalcLLL := 0;
end;

function calcLLLTable( var LLL : TLLL; len, in_low, in_exp, in_high : integer;
                       var out_low, out_high, Chain_Prec : integer) : boolean;
var
  {b,} i, j, {k,} x,
  new_len, new_low, new_high, alt_low, alt_high, new_Chain_Prec : integer;
  new_LLL : TLLL;
  done : boolean;
begin
  //showDebug( LLL, in_low, len, FALSE, '!');
  //showDebug( LLL, in_high, len, TRUE, '!');
  Chain_Prec := -1;
  if len = 2 then
  begin
    out_high := LL[LLL[1],LLL[2]];
    out_low := LL[LLL[1],LLL[2]];
    showDebug( LLL, out_high, 2, TRUE, 'A');
    calcLLLTable := TRUE;
    exit;
  end;

  if LLL[len] - LLL[len - 1] < 3 then
  begin
    done := calcLLLTable( LLL, len - 1, in_low - 1, in_exp - 1, in_high - 1,
                          new_low, new_high, new_Chain_Prec);
    out_low := new_low + 1;
    out_high := new_high + 1;
    Chain_Prec := LLL[len - 1];
    showDebug( LLL, out_high, len, done, 'B');
    calcLLLTable := done;
    exit;
  end;

  j := len - 1;
  while (2 * LLL[j] >= LLL[len]) and ( j > 0) do
  begin
    i := 1;
    while LLL[i] + LLL[j] < LLL[len] do
      inc(i);
    if LLL[i] + LLL[j] = LLL[len] then
    begin
      done := calcLLLTable( LLL, len - 1, in_low - 1, in_exp - 1, in_high - 1,
                            new_low, new_high, new_Chain_Prec);
      out_low := new_low + 1;
      out_high := new_high + 1;
      Chain_Prec := LLL[len - 1];
      showDebug( LLL, out_high, len, done, 'C');
      calcLLLTable := done;
      exit;
    end;
    dec(j);
  end;

  for i := 1 to len - 1 do
    if LL[LLL[i],LLL[len]] > in_low then
    begin
      in_low := LL[LLL[i],LLL[len]];
      if in_low >= in_exp then
      begin
        out_low := in_exp;
        out_high := in_exp;
        showDebug( LLL, out_high, len, FALSE, 'D');
        calcLLLTable := FALSE;
        exit;
      end;
    end;

  for i := 1 to len - 2 do
    for j := i + 1 to len - 1 do
    begin
      if (2 * LLL[j] < LLL[j+1]) AND
         (LL[LLL[i],LLL[j]] + len - j + 1 > in_low) then
      begin
        in_low := LL[LLL[i],LLL[j]] + len - j + 1;
        if in_low >= in_exp then
        begin
          out_low := in_exp;
          out_high := in_exp;
          showDebug( LLL, out_high, len, FALSE, 'E');
          calcLLLTable := FALSE;
          exit;
        end;
      end
      else if LL[LLL[i],LLL[j]] + len - j > in_low then
      begin
        in_low := LL[LLL[i],LLL[j]] + len - j;
        if in_low >= in_exp then
        begin
          out_low := in_exp;
          out_high := in_exp;
          showDebug( LLL, out_high, len, FALSE, 'F');
          calcLLLTable := FALSE;
          exit;
        end;
      end;
    end;

  if (LLL[2] > 2 * LLL[1]) AND (L[LLL[1]] + len > in_low) then
  begin
    in_low := L[LLL[1]] + len;
    if in_low >= in_exp then
    begin
      out_low := in_exp;
      out_high := in_exp;
      showDebug( LLL, out_high, len, FALSE, 'G');
      calcLLLTable := FALSE;
      exit;
    end;
  end;

  for i := 2 to len - 1 do
    for j := 1 to i - 1 do
    begin
      if (LLL[i+1] > 2 * LLL[i]) AND
         (LL[LLL[j],LLL[i]] + len - i + 1 > in_low) then
      begin
        in_low := LL[LLL[j],LLL[i]] + len - i + 1;
        if in_low >= in_exp then
        begin
          out_low := in_exp;
          out_high := in_exp;
          showDebug( LLL, out_high, len, FALSE, 'H');
          calcLLLTable := FALSE;
          exit;
        end;
      end
      else if LL[LLL[j],LLL[i]] + len - i > in_low then
      begin
        in_low := LL[LLL[j],LLL[i]] + len - i;
        if in_low >= in_exp then
        begin
          out_low := in_exp;
          out_high := in_exp;
          showDebug( LLL, out_high, len, FALSE, 'I');
          calcLLLTable := FALSE;
          exit;
        end;
      end;
    end;

  //new_high := in_high;
  new_low := in_high;

  if LLL[len] mod 2 = 0 then
  begin
    i := 1;
    j := 1;
    while LLL[i] < LLL[len] div 2 do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_LLL[j] := LLL[len] div 2;
    inc(j);
    if 2 * LLL[i] = LLL[len] then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;
    Chain_Prec := new_LLL[new_len];
    done := calcLLLTable( new_LLL, new_len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, New_Chain_Prec);
    if (alt_high + 1 <= in_exp) then
    begin
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        out_low := in_low;
        out_high := in_low;
        showDebug( LLL, out_high, len, done, 'L');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;

    for i := 1 to len - 1 do new_LLL[i] := LLL[i];
    new_LLL[len] := LLL[len] - 2;
    done := calcLLLTable( new_LLL, len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, New_Chain_Prec);
    if (alt_high + 1 < in_exp) then
    begin
      Chain_Prec := LLL[len] - 2;
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        out_low := in_low;
        out_high := in_low;
        showDebug( LLL, out_high, len, done, 'K');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;

    new_LLL[len] := LLL[len] - 1;

    done := calcLLLTable( new_LLL, len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, new_Chain_Prec);
    if (alt_high + 1 < in_exp) then
    begin
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        Chain_Prec := LLL[new_len];
        out_low := in_exp;
        out_high := in_exp;
        showDebug( LLL, out_high, len, done, 'J');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;
  end
  else
  begin
    for i := 1 to len - 1 do new_LLL[i] := LLL[i];
    new_LLL[len] := LLL[len] - 1;

    Chain_Prec := LLL[len] - 1;

    done := calcLLLTable( new_LLL, len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, new_Chain_Prec);
    if (alt_high + 1 < in_exp) then
    begin
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        out_low := in_exp;
        out_high := in_exp;
        showDebug( LLL, out_high, len, done, 'J');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;

    new_LLL[len] := LLL[len] - 2;
    done := calcLLLTable( new_LLL, len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, New_Chain_Prec);
    if (alt_high + 1 < in_exp) then
    begin
      Chain_Prec := LLL[len] - 2;
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        out_low := in_low;
        out_high := in_low;
        showDebug( LLL, out_high, len, done, 'K');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;
  end;

  if len >= MAXLLL then showmessage('Out of resources');

  for x := 3 to (LLL[len] - 1) div 2 do
  begin
    i := 1;
    j := 1;
    while LLL[i] < x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_LLL[j] := x;
    inc(j);
    if LLL[i] = x then inc(i);
    while LLL[i] < LLL[len] - x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if 2 * x <> LLL[len] then
    begin
      new_LLL[j] := LLL[len] - x;
      inc(j);
    end;
    if LLL[i] = LLL[len] - x then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;
    done := calcLLLTable( new_LLL, new_len, in_low - 1, in_exp - 1, in_high + 1,
                          alt_low, alt_high, new_Chain_Prec);
    if (alt_high + 1 < in_exp) then
    begin
      Chain_Prec := new_LLL[new_len];
      in_exp := alt_high + 1;
      if in_exp = in_low then
      begin
        out_low := in_low;
        out_high := in_low;
        showDebug( LLL, out_high, len, done, 'M');
        calcLLLTable := done;
        exit;
      end;
    end;
    if alt_low + 1 < new_low then new_low := alt_low + 1;
  end;
  out_high := new_low;
  done := (new_low = in_exp);
  showDebug( LLL, out_high, len, done, 'Q');
  calcLLLTable := done;
end;

function calcLL(x, y, in_low, in_high : integer) : integer;
var
  len, Chain_Prec : integer;
  LLL : TLLL;
  txt : string;
begin
  if y <= 4 then
  begin
    chain[0] := 1;
    chain[1] := 2;
    len := 3;
    if x > 2 then chain[2] := x else dec(len);
    if y > 2 then chain[len] := y else dec(len);
    calcLL := len;
    exit;
  end;
  Chain_Prec := ChainData[x,y];
  if (Chain_Prec = x) OR (Chain_Prec + x = y) OR (2 * Chain_Prec = y) OR
     (Chain_Prec + 1 = y) OR (Chain_Prec + 2 = y) then
  begin
    if (2 * Chain_Prec = y) AND (2 * x < y)
      then len := calcLL(x, y div 2 , in_low - 1, in_high - 1)
    else if (2 * Chain_Prec = y) AND (2 * x > y)
      then len := calcLL(y div 2, x , in_low - 1, in_high - 1)
    else if (2 * Chain_Prec = y) AND (2 * x = y)
      then len := calcLL(1, x, in_low - 1, in_high - 1)
    else if Chain_Prec = x
      then len := calcLL(y - x, x, in_low - 1, in_high - 1)
    else if Chain_Prec + x = y
      then len := calcLL(x, y - x, in_low - 1, in_high - 1)
    else if Chain_Prec + 1 = y
      then len := calcLL(x, y - 1, in_low - 1, in_high - 1)
      else len := calcLL(x, y - 2, in_low - 1, in_high - 1);
    inc(len);
    chain[len] := y;
    calcLL := len;
    exit;
  end;
  if x < y - Chain_Prec then LLL[1] := x else LLL[1] := y - Chain_Prec;
  if x > Chain_Prec then LLL[3] := x else LLL[3] := Chain_Prec;
  LLL[2] := x + y - LLL[1] - LLL[3];
  if in_low = in_high then txt := 'Z' else txt := '?';
  showDebug( LLL, in_high - 1, 3, TRUE, txt);
  len := calcLLL(LLL, 3, in_low - 1, in_high - 1);
  inc(len);
  chain[len] := y;
  calcLL := len;
end;

procedure CalcLLTabel;
var
  x, xx, y, LL_low, LL_high, LL_altlow, LL_althigh, LL_newlow,
  LL_diff, Chain_Prec, XChain_Prec : integer;
  LLL : TLLL;
  done : boolean;
begin
  maxLL := TableU.F_Table.S_LL.value;
  for x := 1 to maxLL do
  for y := 1 to maxLL do
  begin
    LL[x,y] := 0;
    LLdiff[x,y] := 0;
    LLbrun[x,y] := 0;
  end;

  L[1] := 0;
  L[2] := 1;
  LL[1,2] := 1;

  for y := 3 to maxLL do
  begin
    L[y] := L[y-1] + 1;
    ChainData[y,y] := y - 1;
    if (y mod 2 = 0) AND (L[y div 2] + 1 < L[y]) then
    begin
      L[y] := L[y div 2] + 1;
      ChainData[y,y] := y div 2;
    end;
    for x := 3 to (y - 1) div 2 - 1 do
    begin
      if LL[x,y-x] + 1 < L[y] then
      begin
        L[y] := LL[x,y-x] + 1;
        ChainData[y,y] := y - x;
      end;
    end;
    LL[y-2,y] := L[y-2] + 1;
    ChainData[y-2,y] := y - 2;
    LL[y-1,y] := L[y-1] + 1;
    ChainData[y-1,y] := y - 1;
    LL[1,y] := L[y];
    ChainData[1,y] := ChainData[y,y];
    LL[2,y] := L[y];
    ChainData[2,y] := ChainData[y,y];
    if y mod 2 = 0 then
    begin
      LL[y div 2, y] := L[y div 2] + 1;
      ChainData[y div 2,y] := y div 2;
    end;
    for x := 3 to y - 3 do
    if 2 * x <> y then
    begin
      LL_low := L[y];
      LL_high := LL[x,y div 2] + LL[y div 2,x] + 1;
      if L[x] + 1 > LL_low then LL_low := L[x] + 1;
      if (2 * x > y) AND (L[y-x] + 2 > LL_low) then LL_low := L[y-x] + 2;
      if y mod 2 = 0 then
      begin
        if 2 * x < y then ChainData[x,y] := y div 2 else ChainData[x,y] := x;
        LL_althigh := LL[x,y div 2] + LL[y div 2,x] + 1;
        if LL_althigh < LL_high then
        begin
          LL_high := LL_althigh;
          ChainData[x,y] := y div 2;
        end;

        if LL[x,y-2] + 1 < LL_high then
        begin
          LL_high := LL[x, y-2] + 1;
          ChainData[x,y] := y - 2;
        end;

        if LL[x,y-x] + LL[y-x,x] + 1 < LL_high then
        begin
          LL_high := LL[x,y-x] + LL[y-x,x] + 1;
          if 2 * x < y then ChainData[x,y] := y - x
                       else ChainData[x,y] := x;
        end;
      end
      else
      begin
        LL_high := LL[x, y-1] + 1;
        ChainData[x,y] := y - 1;
        if LL[x,y-x] + LL[y-x,x] + 1 < LL_high then
        begin
          LL_high := LL[x,y-x] + LL[y-x,x] + 1;
          if 2 * x < y then ChainData[x,y] := y - x
                       else ChainData[x,y] := x;
        end;

        if LL[x,y-2] + 1 < LL_high then
        begin
          LL_high := LL[x, y-2] + 1;
          ChainData[x,y] := y - 2;
        end;
      end;

      LL_newlow := LL_high;
      if LL_low < LL_high then
      begin
        for xx := 3 to (y - 1) div 2 do
        begin
        if (xx <> x) AND (x + xx <> y) AND (LL_low < LL_high) AND
           (LL[xx,y-xx] + 1 < LL_high) AND
           (LL[x,y-xx] + LL[y-xx,x] + 1 < LL_high) then
        begin
          Chain_Prec := y - xx;
          if (xx < x) AND (2 * xx < y) then LLL[1] := xx;
          if (xx < x) AND (2 * xx > y) then LLL[1] := y - xx;
          if (xx > x) AND (x + xx < y) then LLL[1] := x;
          if (xx > x) AND (x + xx > y) then LLL[1] := y - xx;
          if (xx > x) AND (2 * xx > y) then LLL[3] := xx;
          if (xx > x) AND (2 * xx < y) then LLL[3] := y - xx;
          if (xx < x) AND (x + xx > y) then LLL[3] := x;
          if (xx < x) AND (x + xx < y) then LLL[3] := y - xx;
          LLL[2] := y + x - LLL[1] - LLL[3];
          DebugU.F_Debug.M_Debug.Lines.Add( '!(' + inttostr(x) + ',' +
                                            inttostr(y) +') [' + inttostr(xx) +
                                            '] ' + inttostr(LL_low) +
                                            ' <= ... <= ' + inttostr(LL_high));
          done := calcLLLTable( LLL, 3, LL_low - 1, LL_high - 1, LL_high - 1,
                                LL_altlow, LL_althigh, XChain_Prec);
          //if LL_altlow = LL_althigh then
          ChainData[x,y] := Chain_Prec;
          if done then
          DebugU.F_Debug.M_Debug.Lines.Add( '+>>L(' + inttostr(LLL[1]) + ',' +
                                            inttostr(LLL[2]) + ',' +
                                            inttostr(LLL[3]) + ',' + inttostr(y)
                                            + ') = ' + inttostr(LL_altlow+1) +
                                            ' >> ' + inttostr(Chain_Prec))
          else
          DebugU.F_Debug.M_Debug.Lines.Add( '-->>L(' + inttostr(LLL[1]) + ',' +
                                            inttostr(LLL[2]) + ',' +
                                            inttostr(LLL[3]) + ',' + inttostr(y)
                                            + ') = [' + inttostr(LL_altlow+1)
                                            + '/' + inttostr(LL_althigh+1) + ']');
          if LL_althigh + 1 < LL_high then LL_high := LL_althigh + 1;
          if LL_altlow + 1 < LL_newlow then LL_newlow := LL_altlow + 1;
        end;
        end;
      end;
      LL[x,y] := LL_high;
      LLdiff[x,y] := LL_high - LL_newlow;
    end;
  end;
end;

procedure TF_Chain.FormCreate(Sender: TObject);
begin
  maxLL := 0;
end;

procedure TF_Chain.FormActivate(Sender: TObject);
begin
  //F_Debug.Show;
end;

procedure TF_Chain.B_CalcClick(Sender: TObject);
var
  len, x, y, i : integer;
  s : string;
begin
  if S_N.value > maxLL then
  begin
    ShowMessage('The table must be made larger!');
    exit;
  end;
  x := S_M.value;
  y := S_N.value;
  if x >= y then exit;
  len := calcLL(x, y, LL[x,y] - LLdiff[x,y], LL[x,y]);
  s := '(';
  for i := 0 to len - 1 do
    s := s + inttostr(chain[i]) + ',';
  s := s + inttostr(chain[len]) + ')';
  E_Chain.Text := s;
  E_Len.Text := inttostr(len);
end;

procedure TF_chain.B_TableClick(Sender: TObject);
begin
  F_Table.ShowModal;
end;

procedure TF_Chain.S_Exit(Sender: TObject);
var
  chnge : integer;
begin
  if S_M.value >= S_N.value then
  begin
    chnge := S_M.value;
    S_M.value := S_N.value;
    S_N.value := chnge;
    showmessage('M must be smaller than N.');
  end;
end;

procedure TF_Chain.S_Change(Sender: TObject);
begin
  E_Chain.Text := '()';
  E_Len.Text := '0';
end;

procedure TF_Chain.B_DebugClick(Sender: TObject);
begin
  F_Debug.Show;
end;

procedure TF_Chain.B_ExitClick(Sender: TObject);
begin
  Close;
end;

end.
