unit ChainU;

{
  Dit programma bepaalt de exacte chain-lengte voor addition-chains voor een en
  twee elementen. Ook is het mogelijk om een exacte chain te laten creeren.

  We onderscheiden triviale en niet triviale chains. De triviale chains worden
  door het Brun-algorithme bepaald. De niet-triviale chains worden opgeslagen.
}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Spin, StdCtrls;

const
  MAXLLL = 15;
  MAXchain = 20;
  MAXL = 248;

type
  TLLL = array[1..MAXLLL] of longint;
  Tchain = array[0..MAXchain] of longint;
  TL = array[1..MAXL] of integer;
  TLL = array[1..MAXL,1..MAXL] of integer;

  TF_Chain = class(TForm)
    B_Calc: TButton;
    B_Exit: TButton;
    E_Chain: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    S_M: TSpinEdit;
    S_N: TSpinEdit;
    Label4: TLabel;
    E_Len: TEdit;
    B_Table: TButton;
    B_Debug: TButton;
    procedure B_CalcClick(Sender: TObject);
    procedure B_ExitClick(Sender: TObject);
    procedure B_TableClick(Sender: TObject);
    procedure S_Exit(Sender: TObject);
    procedure S_Change(Sender: TObject);
    procedure B_DebugClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

procedure CalcLLTabel;
function calcLL(x, y, DDin, LLin : integer) : integer;
function calcLLBrun(x, y : integer) : integer;
function calcLLModBrun(x, y : integer) : integer;
procedure calcLLLTable( var LLL : TLLL; len : integer;
                        var DDvalue, LLvalue, Chain_Prec : integer);

var
  F_Chain : TF_Chain;
  Chain : TChain;
  L : TL;
  DD, LL, LLbrun, ChainData : TLL;
  maxLL, cnt, status : integer;
  
implementation

uses TableU, DebugU;

{$R *.DFM}

function ChainDataNotation(x,y:integer):integer;
var new_x : integer;
begin
  if 2 * x < y then new_x := x else new_x := y - x;
  if (y mod new_x = 0) AND (y div new_x < 10)
    then ChainDataNotation := -y div new_x
    else ChainDataNotation := new_x;
end;

procedure showDebugLLL( var LLL : TLLL; len, DDvalue, LLvalue : integer;
                        txt : string);
var
  i : integer;
  s, t : string;
begin
  if (pos(txt, DebugU.S_Debug) = 0) OR
     (LLL[len] < DebugU.F_Debug.S_Min.value) then exit;
  if DDvalue = LLvalue then s := txt + ':      L('
                       else s := txt + ': ' + inttostr(ddValue) + ' <= L(';
  t := '';
  for i := 1 to len - 1 do t := t + inttostr(LLL[i]) + ',';
  if DDvalue = LLvalue
    then s := s + t + inttostr(LLL[len]) + ') = ' + inttostr(LLvalue)
    else s := s + t + inttostr(LLL[len]) + ') <= ' + inttostr(LLvalue);
  DebugU.F_Debug.M_Debug.Lines.Add(s);
  Application.ProcessMessages;
end;




procedure showDebugLL( x, y, DDvalue, LLvalue : integer; txt : string);
var
  s : string;
begin
  if (pos(txt, DebugU.S_Debug) = 0) OR
     (y < DebugU.F_Debug.S_Min.value) then exit;
  if DDvalue = LLvalue
    then s := txt + ':      L('
    else s := txt + ': ' + inttostr(ddValue) + ' <= L(';
  s := s + inttostr(x) + ',' + inttostr(y);
  if DDvalue = LLvalue
    then s := s + ') = ' + inttostr(LLvalue)
    else s := s + ') <= ' + inttostr(LLvalue);
  DebugU.F_Debug.M_Debug.Lines.Add(s);
  Application.ProcessMessages;
end;




function calcLLL( var LLL : TLLL; len, DDin, LLin : integer) : integer;
var
  DDvalue, LLvalue, new_len, i, j, x, q, Lr, Chain_Prec : integer;
  new_LLL : TLLL;
  new_chain : TChain;
begin
  q := 1;
  DDValue := DDin;
  LLValue := LLin;
  if (LLvalue <= 0) OR (DDvalue <= 0) then
  begin
    showDebugLLL( LLL, len, DDvalue, LLvalue, 'k');
    if DDvalue <= 0 then
    begin
      DebugU.F_Debug.M_Debug.Lines.Add('cannot apply calcLLL on 0');
      calcLLL := 0;
      exit;
    end;
    LLvalue := DDvalue;
  end;
  if len = 2 then
  begin
    calcLLL := calcLL(LLL[1], LLL[2], DDin, LLin);
    exit;
  end;
  if LLL[1] < 3 then
  begin
    for i := 1 to len - 1 do new_LLL[i] := LLL[i+1];
    Lr := calcLLL(new_LLL, len - 1, DDin, LLin);
    calcLLL := Lr;
    exit;
  end;

  if LLL[len] - LLL[len-1] < 3 then
  begin
    Lr := calcLLL(LLL, len - 1, DDin - 1, LLin - 1);
    inc(Lr);
    chain[Lr] := LLL[len];
    calcLLL := Lr;
    exit;
  end;

  j := len - 1;
  while (2 * LLL[j] >= LLL[len]) and ( j > 0) do
  begin
    i := 1;
    while LLL[i] + LLL[j] < LLL[len] do
      inc(i);
    if LLL[i] + LLL[j] = LLL[len] then
    begin
      Lr := calcLLL(LLL, len - 1, DDvalue - 1, LLvalue - 1);
      inc(Lr);
      chain[Lr] := LLL[len];
      calcLLL := Lr;
      exit;
    end;
    dec(j);
  end;
  //{toevoeging (temp) > toch wel vrij definitief}
  //inc(LLvalue);
  calcLLLTable( LLL, len, DDvalue, LLvalue, Chain_Prec);
  if (LLvalue <= 0) OR (DDvalue <= 0) OR (LLvalue < DDValue) then
  begin
    showDebugLLL( LLL, len, DDvalue, LLvalue, 'e');
    LLvalue := DDvalue;
    //if DDvalue <= 0 then
    begin
      DebugU.F_Debug.M_Debug.Lines.Add('cannot continue calcLLL because of bad values');
      calcLLL := DDvalue;
      exit;
    end;
  end;
  showDebugLLL( LLL, len, DDvalue, LLvalue, 'a');
  if Chain_Prec = 0 then
  begin
    q := LLL[len] div LLL[len - 1];
    new_len := calcLL(1, q, L[q], L[q]);
    for i := 1 to new_len do
      new_chain[i] := LLL[len-1] * chain[i];
    new_len := len - 1;
    for i := 1 to new_len do new_LLL[i] := LLL[i];
  end
  else
  begin
    if 2 * Chain_Prec < LLL[len] then Chain_Prec := LLL[len] - Chain_Prec;
    x := LLL[len] - Chain_Prec;
    i := 1;
    j := 1;
    while (LLL[i] < x) AND (i < len) do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if x > 2 then
    begin
      new_LLL[j] := x;
      inc(j);
    end;
    if LLL[i] = x then inc(i);
    while (LLL[i] < LLL[len] - x) AND (i < len) do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if 2 * x <> LLL[len] then
    begin
      new_LLL[j] := LLL[len] - x;
      inc(j);
    end;
    if LLL[i] = LLL[len] - x then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;
  end;
  DebugU.F_Debug.M_Debug.Lines.Add('------------------------------------------');
  Lr := calcLLL(new_LLL, new_len, DDvalue - 1, LLvalue);
  if Chain_Prec = 0 then
  begin
    for i := 1 to L[q] do chain[Lr + i] := new_chain[i];
    Lr := Lr + L[q];
  end
  else
  begin
    inc(Lr);
    chain[Lr] := LLL[len];
  end;
  calcLLL := Lr;
  if DDvalue <> LLvalue then showmessage('Unsure!');
end;




procedure calcLLLTable( var LLL : TLLL; len : integer;
                        var DDvalue, LLvalue, Chain_Prec : integer);
var
  i, j, x, DDnew, LLnew, DDDout, LLLin, new_len, DUMMY : integer;
  new_LLL : TLLL;
  s, t : string;
begin
  if len >= MAXLLL then
  begin
    s := 'Out of resources for length ' + inttostr(len) + '(';
    t := '';
    for i := 1 to len - 1 do t := t + inttostr(LLL[i]) + ',';
    s := s + t + inttostr(LLL[len]) + ')';
    DebugU.F_Debug.M_Debug.Lines.Add(s);
    Application.ProcessMessages;
    DDvalue := 0;
    LLvalue := 0;
    exit;
  end;




  showDebugLLL( LLL, len, DDvalue, LLvalue, 'I');

  if (DDvalue > LLvalue) OR (LLvalue <= 0) then
  begin
    s := '! ' + inttostr(len) + ' : (';
    t := '';
    for i := 1 to len - 1 do t := t + inttostr(LLL[i]) + ',';
    s := s + t + inttostr(LLL[len]) + ')  >> exit...';
    DebugU.F_Debug.M_Debug.Lines.Add(s);
    Application.ProcessMessages;
    DDvalue := 0;
    LLvalue := 0;
    exit;
  end;

  Chain_Prec := -1;
  if len = 2 then
  begin
    DDvalue := LL[LLL[1],LLL[2]];
    LLvalue := LL[LLL[1],LLL[2]];
    Chain_Prec := ChainData[LLL[1],LLL[2]];
    showDebugLLL( LLL, 2, DDvalue, LLvalue, 'A');
    exit;
  end;

  for i := 1 to len - 1 do
    if LL[LLL[i],LLL[len]] > DDvalue then
    begin
      DDvalue := LL[LLL[i],LLL[len]];
      if DDvalue >= LLvalue then
      begin
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'B');
        exit;
      end;
    end;

  if LLL[len] - LLL[len - 1] < 3 then
  begin
    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( LLL, len - 1, DDnew, LLnew, DUMMY);
    DDvalue := DDnew + 1;
    LLvalue := LLnew + 1;
    Chain_Prec := LLL[len - 1];
    showDebugLLL( LLL, len, DDvalue, LLvalue, 'C');
    exit;
  end;

  if LLL[len] mod LLL[len - 1] = 0 then
  begin
    x := L[LLL[len] div LLL[len - 1]];
    DDnew := DDvalue - x;
    LLnew := LLvalue - x;
    calcLLLTable( LLL, len - 1, DDnew, LLnew, DUMMY);
    if (LLnew + x < LLvalue) then
    begin
      LLvalue := LLnew + x;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := 0;
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'D');
        exit;
      end;
    end;
  end;

  j := len - 1;
  while (2 * LLL[j] >= LLL[len]) and ( j > 0) do
  begin
    i := 1;
    while LLL[i] + LLL[j] < LLL[len] do
      inc(i);
    if LLL[i] + LLL[j] = LLL[len] then
    begin
      DDnew := DDvalue - 1;
      LLnew := LLvalue - 1;
      calcLLLTable( LLL, len - 1, DDnew, LLnew, DUMMY);
      DDvalue := DDnew + 1;
      LLvalue := LLnew + 1;
      Chain_Prec := LLL[len - 1];
      showDebugLLL( LLL, len, DDvalue, LLvalue, 'E');
      exit;
    end;
    dec(j);
  end;

  for i := 1 to len - 2 do
    for j := i + 1 to len - 1 do
    begin
      if (2 * LLL[j] < LLL[j+1]) AND
         (LL[LLL[i],LLL[j]] + len - j + 1 > DDvalue) then
      begin
        DDvalue := LL[LLL[i],LLL[j]] + len - j + 1;
        if DDvalue >= LLvalue then
        begin
          showDebugLLL( LLL, len, DDvalue, LLvalue, 'F');
          exit;
        end;
      end
      else if LL[LLL[i],LLL[j]] + len - j > DDvalue then
      begin
        DDvalue := LL[LLL[i],LLL[j]] + len - j;
        if DDvalue >= LLvalue then
        begin
          showDebugLLL( LLL, len, DDvalue, LLvalue, 'G');
          exit;
        end;
      end;
    end; {double for}

  if (LLL[2] > 2 * LLL[1]) AND (L[LLL[1]] + len > DDvalue) then
  begin
    DDvalue := L[LLL[1]] + len;
    if DDvalue >= LLvalue then
    begin
      showDebugLLL( LLL, len, DDvalue, LLvalue, 'H');
      exit;
    end;
  end;

  for i := 2 to len - 1 do
    for j := 1 to i - 1 do
    begin
      if (LLL[i+1] > 2 * LLL[i]) AND
         (LL[LLL[j],LLL[i]] + len - i + 1 > DDvalue) then
      begin
        DDvalue := LL[LLL[j],LLL[i]] + len - i + 1;
        if DDvalue >= LLvalue then
        begin
          showDebugLLL( LLL, len, DDvalue, LLvalue, 'J');
          exit;
        end;
      end
      else if LL[LLL[j],LLL[i]] + len - i > DDvalue then
      begin
        DDvalue := LL[LLL[j],LLL[i]] + len - i;
        if DDvalue >= LLvalue then
        begin
          showDebugLLL( LLL, len, DDvalue, LLvalue, 'K');
          exit;
        end;
      end;
    end; {double for}

  if LLL[len] mod 2 = 0 then
  begin
    i := 1;
    j := 1;
    while 2 * LLL[i] < LLL[len] do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_LLL[j] := LLL[len] div 2;
    inc(j);
    if 2 * LLL[i] = LLL[len] then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;
    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, new_len, DDnew, LLnew, DUMMY);
    if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := LLL[len] div 2;
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'L');
        exit;
      end;
    end;

    for i := 1 to len - 1 do new_LLL[i] := LLL[i];
    new_LLL[len] := LLL[len] - 2;
    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, len, DDnew, LLnew, DUMMY);
    if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := LLL[len] - 2;
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'M');
        exit;
      end;
    end;

    new_LLL[len] := LLL[len] - 1;

    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, len, DDnew, LLnew, DUMMY);
    if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := LLL[len] - 1;
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'N');
        exit;
      end;
    end;
  end
  else
  begin
    for i := 1 to len - 1 do new_LLL[i] := LLL[i];
    new_LLL[len] := LLL[len] - 1;

    Chain_Prec := LLL[len] - 1;

    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, len, DDnew, LLnew, DUMMY);
    if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := LLL[len - 1];
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'n');
        exit;
      end;
    end;

    new_LLL[len] := LLL[len] - 2;
    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, len, DDnew, LLnew, DUMMY);
     if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := LLL[len] - 2;
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'm');
        exit;
      end;
    end;
  end;

  for x := 3 to (LLL[len] - 1) div 2 do
  begin

    i := 1;
    j := 1;
    while LLL[i] < x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_LLL[j] := x;
    inc(j);
    if LLL[i] = x then inc(i);
    while LLL[i] < LLL[len] - x do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    if 2 * x <> LLL[len] then
    begin
      new_LLL[j] := LLL[len] - x;
      inc(j);
    end;
    if LLL[i] = LLL[len] - x then inc(i);
    while i < len do
    begin
      new_LLL[j] := LLL[i];
      inc(i);
      inc(j);
    end;
    new_len := j - 1;

    DDnew := DDvalue - 1;
    LLnew := LLvalue - 1;
    calcLLLTable( new_LLL, new_len, DDnew, LLnew, DUMMY);

    if (LLnew + 1 < LLvalue) then
    begin
      LLvalue := LLnew + 1;
      if LLvalue = DDvalue then
      begin
        Chain_Prec := new_LLL[new_len];
        showDebugLLL( LLL, len, DDvalue, LLvalue, 'Q');
        exit;
      end;
    end;
  end;
  if LLvalue <> DDvalue then inc(DDvalue);
  showDebugLLL( LLL, len, DDvalue, LLvalue, 'q');
end;




function calcLLBrun(x, y : integer) : integer;
begin
  if y < 4 then
  begin
    calcLLBrun := y - 1;
    exit;
  end;
  if y mod x = 0 then
  begin
    calcLLBrun := L[y] + L[y div x];
    exit;
  end;
  calcLLBrun := calcLLBrun(y mod x, x) + L[y div x] + 1;
end;




function calcLLModBrun(x, y : integer) : integer;
begin
  if y < 4 then
  begin
    calcLLModBrun := y - 1;
    exit;
  end;
  if y mod x = 0 then
  begin
    calcLLModBrun := L[x] + L[y div x];
    exit;
  end;
  calcLLModBrun := LL[y mod x,x] + L[y div x] + 1;
end;




function calcLL(x, y, DDin, LLin : integer) : integer;
var
  i, len, higherlen, Chain_Prec : integer;
  LLL : TLLL;
  txt : string;
  higherchain : TChain;
begin
  if (LLin <= 0) OR (DDin <= 0) then
  begin
    DebugU.F_Debug.M_Debug.Lines.Add('cannot apply calcLL on 0');
    calcLL := 0;
    exit;
  end;
  showDebugLL( x, y, DDin, LLin, 'g');
  if cnt > 10 * LLin then begin calcLL := 0;exit end
              else inc(cnt);
  if y <= 1 then
  begin
    calcLL := 0;
    exit;
  end;
  if y <= 4 then
  begin
    chain[0] := 1;
    chain[1] := 2;
    len := 3;
    if x > 2 then chain[2] := x else dec(len);
    if y > 2 then chain[len] := y else dec(len);
    calcLL := len;
    exit;
  end;
  if (x = 1) AND (ChainData[y,y] = 0) then x := 2;
  if ChainData[x,y] = 0 then
  begin
    higherlen := calcLL(1, y div x, L[y div x], L[y div x]+1);
    for i := 1 to higherlen do higherchain[i] := chain[i];
    if y mod x = 0 then
    begin
      len := calcLL(1, x, L[x], L[x]+1);
      for i := 1 to higherlen do chain[len+i] := x * higherchain[i];
      len := len + higherlen;
    end
    else
    begin
      len := calcLL(y mod x, x, L[x], L[x]+1);
      for i := 1 to higherlen do chain[len+i] := x * higherchain[i];
      chain[len+higherlen+1] := y;
      len := len + higherlen + 1;
    end;
    calcLL := len;
    exit;
  end;
  if 2 * ChainData[x,y] > y then showmessage('ERROR in DATA');
  if ChainData[x,y] < 0 then Chain_Prec := y + y div ChainData[x,y]
                         else Chain_Prec := y - ChainData[x,y];
  if (Chain_Prec = x) OR (Chain_Prec + x = y) OR (2 * Chain_Prec = y) OR
     (Chain_Prec + 1 = y) OR (Chain_Prec + 2 = y) then
  begin
  //showDebugLL( x, y, DDin, LLin, 'B');
    if (2 * Chain_Prec = y) AND (2 * x < y)
      then len := calcLL(x, y div 2 , DDin - 1, LLin - 1)
    else if (2 * Chain_Prec = y) AND (2 * x > y)
      then len := calcLL(y div 2, x , DDin - 1, LLin - 1)
    else if (2 * Chain_Prec = y) AND (2 * x = y)
      then len := calcLL(1, x, DDin - 1, LLin - 1)
    else if Chain_Prec = x
      then len := calcLL(y - x, x, DDin - 1, LLin - 1)
    else if Chain_Prec + x = y
      then len := calcLL(x, y - x, DDin - 1, LLin - 1)
    else if Chain_Prec + 1 = y
      then len := calcLL(x, y - 1, DDin - 1, LLin - 1)
      else len := calcLL(x, y - 2, DDin - 1, LLin - 1);
  //showDebugLL( x, y, DDin, LLin, 'C');
    inc(len);
    chain[len] := y;
    calcLL := len;
    exit;
  end;
  //showDebugLL( x, y, DDin, LLin, 'D');
  if x < y - Chain_Prec then LLL[1] := x else LLL[1] := y - Chain_Prec;
  if x > Chain_Prec then LLL[3] := x else LLL[3] := Chain_Prec;
  LLL[2] := x + y - LLL[1] - LLL[3];
  if DDin = LLin then txt := 'h' else txt := '?';
  showDebugLLL( LLL, 3, DDin - 1, LLin - 1, txt);
  len := calcLLL(LLL, 3, DDin - 1, LLin - 1);
  showDebugLLL( LLL, 3, len, len, 'd');
  inc(len);
  chain[len] := y;
  calcLL := len;
end;

procedure CalcLLTabel;
var
  k, x, xx, y, DDD, DDvalue, LLvalue, Chain_Prec, DUMMY : integer;
  LLL : TLLL;
  pref : Array[1..MAXL div 2] of integer;
begin
  for x := 1 to maxLL do
  for y := 1 to maxLL do
  begin
    LL[x,y] := 0;
    DD[x,y] := 0;
    ChainData[x,y] := 0;
    LLbrun[x,y] := 0;
  end;

  L[1] := 0;
  L[2] := 1;
  DD[1,2] := 1;
  LL[1,2] := 1;

  for y := 3 to maxLL do
  begin
    if y mod 2 = 0 then
    begin
      L[y] := L[y div 2] + 1;
      if L[y-2] + 1 < L[y] then
      begin
        L[y] := L[y-2] + 1;
        ChainData[y,y] := 2;
        ChainData[1,y] := 2;
        ChainData[2,y] := 2;
      end;
      if L[y-1] + 1 < L[y] then
      begin
        L[y] := L[y-1] + 1;
        ChainData[y,y] := 1;
        ChainData[1,y] := 1;
        ChainData[2,y] := 1;
      end;
    end
    else
    begin
      L[y] := L[y-1] + 1;
      if (L[y-2] + 1 < L[y]) AND (y > 3) then
      begin
        L[y] := L[y-2] + 1;
        ChainData[y,y] := 2;
        ChainData[1,y] := 2;
        ChainData[2,y] := 2;
      end;
    end;
    {preferencelist}
    for k := 1 to (y - 1) div 2 - 2 do pref[k] := k + 2;
    if y mod 7 = 0 then
    begin
      for k := y div 7 - 2 downto 2 do pref[k] := pref[k-1];
      pref[1] := y div 7;
    end;
    if y mod 5 = 0 then
    begin
      for k := y div 5 - 2 downto 2 do pref[k] := pref[k-1];
      pref[1] := y div 5;
    end;
    if y mod 3 = 0 then
    begin
      for k := y div 3 - 2 downto 2 do pref[k] := pref[k-1];
      pref[1] := y div 3;
    end;

    for k := 1 to (y - 1) div 2 - 2 do
    begin
      x := pref[k];
      if LL[x,y-x] + 1 < L[y] then
      begin
        L[y] := LL[x,y-x] + 1;
        xx := chaindatanotation(x,y);
        ChainData[y,y] := xx;
        ChainData[1,y] := xx;
        ChainData[2,y] := xx;
      end;
    end;
    LL[y-2,y] := L[y-2] + 1;
    DD[y-2,y] := L[y-2] + 1;
    ChainData[y-2,y] := 0;
    LL[y-1,y] := L[y-1] + 1;
    DD[y-1,y] := L[y-1] + 1;
    ChainData[y-1,y] := 0;
    LL[1,y] := L[y];
    DD[1,y] := L[y];
    LL[2,y] := L[y];
    DD[2,y] := L[y];
    if y mod 2 = 0 then
    begin
      LL[y div 2, y] := L[y div 2] + 1;
      DD[y div 2, y] := L[y div 2] + 1;
      ChainData[y div 2,y] := 0;
    end;
    for x := 3 to y - 3 do
    if 2 * x <> y then
    begin
      status := 1;
      DD[x,y] := L[y];
      ChainData[x,y] := 0;
      LL[x,y] := calcLLModBrun(x, y);
      if y mod 2 = 0 then
      begin
        if LL[x,y div 2] + LL[y div 2,x] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y div 2] + LL[y div 2,x] + 1;
          ChainData[x,y] := -2;
        end;

        if L[x] + 1 > DD[x,y] then DD[x,y] := L[x] + 1;

        if LL[x,y-2] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-2] + 1;
          ChainData[x,y] := 2;
        end;

        if LL[x,y-x] + LL[y-x,x] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-x] + LL[y-x,x] + 1;
          ChainData[x,y] := ChainDataNotation(x,y);
        end;

        if LL[x,y-1] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-1] + 1;
          ChainData[x,y] := 1;
        end;
        //if (10 * x < 3 * y) AND (L[x] + 3 > DD[x,y]) then DD[x,y] := L[x] + 3;
        if (4 * x < y) AND (L[x] + 3 > DD[x,y]) then DD[x,y] := L[x] + 3;
        if (2 * x < y) AND (L[x] + 2 > DD[x,y]) then DD[x,y] := L[x] + 2;
      end
      else
      begin
        if LL[x,y-1] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-1] + 1;
          ChainData[x,y] := 1;
        end;

        if LL[x,y-x] + LL[y-x,x] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-x] + LL[y-x,x] + 1;
          ChainData[x,y] := ChainDataNotation(x,y);
        end;

        if LL[x,y-2] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-2] + 1;
          ChainData[x,y] := 2;
        end;

        //if (10 * x < 3 * y) AND (4 * x <> y) AND (L[x] + 3 > DD[x,y]) then DD[x,y] := L[x] + 3;
        if (4 * x < y) AND (L[x] + 3 > DD[x,y]) then DD[x,y] := L[x] + 3;
        if (2 * x < y) AND (L[x] + 2 > DD[x,y]) then DD[x,y] := L[x] + 2;
        if L[x] + 1 > DD[x,y] then DD[x,y] := L[x] + 1;

        if LL[x,y-x] + LL[y-x,x] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x,y-x] + LL[y-x,x] + 1;
          ChainData[x,y] := ChainDataNotation(x,y);
        end;

        if LL[x,y-2] + 1 < LL[x,y] then
        begin
          LL[x,y] := LL[x, y-2] + 1;
          ChainData[x,y] := 2;
        end;
      end;

      if DD[x,y] < LL[x,y] then
      begin
        for k := 1 to (y - 1) div 2 - 2 do
        begin
          xx := pref[k];
          if (xx <> x) AND (x + xx <> y) AND (DD[x,y] < LL[x,y]) AND
             (LL[xx,y-xx] + 1 < LL[x,y]) AND
             (LL[x,y-xx] + LL[y-xx,x] + 1 < LL[x,y]) then
          begin
            Chain_Prec := y - xx;
            if (xx < x) AND (2 * xx < y) then LLL[1] := xx;
            if (xx < x) AND (2 * xx > y) then LLL[1] := y - xx;
            if (xx > x) AND (x + xx < y) then LLL[1] := x;
            if (xx > x) AND (x + xx > y) then LLL[1] := y - xx;
            if (xx > x) AND (2 * xx > y) then LLL[3] := xx;
            if (xx > x) AND (2 * xx < y) then LLL[3] := y - xx;
            if (xx < x) AND (x + xx > y) then LLL[3] := x;
            if (xx < x) AND (x + xx < y) then LLL[3] := y - xx;
            LLL[2] := y + x - LLL[1] - LLL[3];
            DDvalue := DD[x,y] - 1;
            LLvalue := LL[x,y] - 1;
            showDebugLL(x, y, DDvalue + 1, LLvalue + 1, 'o');
            calcLLLTable( LLL, 3, DDvalue, LLvalue, DUMMY);
            if LLvalue + 1 < LL[x,y] then
            begin
              ChainData[x,y] := ChainDataNotation(Chain_Prec,y);
              LL[x,y] := LLvalue + 1;
            end;
          end;
        end; {For xx}
        if DD[x,y] < LL[x,y] then inc(DD[x,y]);
      end;
      if LL[x,y] < L[y] then exit;
    end; {For x}
  end; {For y}
end;


procedure TF_Chain.FormActivate(Sender: TObject);
var
  F: TextFile;
  x, y, x_width : integer;
  c : char;
  s, t : string;
begin
  {$I-}
  AssignFile(F, 'addition.dat');
  FileMode := 0;
  {$I+}
  if (IOResult <> 0) then
  begin
    maxLL := 4;
    exit;
  end;
  TableU.F_Table.M_Table.Clear;
  Reset(F);
  Readln(F, maxLL);
  TableU.F_Table.S_LL.value := maxLL;
  for y := 1 to maxLL do
  begin
    Read(F, x, c);
    if x <> y then
    begin
      showmessage('1:Wrong structure at '+inttostr(y)+', '+ inttostr(x)+' instead of '+inttostr(y));
      exit;
    end;
    while c = ' ' do Read(F, c);
    if c <> '>' then
    begin
      showmessage('2:Wrong structure at '+inttostr(y)+', #'+ inttostr(ord(c))+' instead of <space>');
      exit;
    end;
    Read(F, L[y], c);
    while c = ' ' do Read(F, c);
    if c <> '>' then
    begin
      showmessage('3:Wrong structure at '+inttostr(y)+', #'+ inttostr(ord(c))+' instead of <space>');
      exit;
    end;
    LL[1,y] := L[y];
    LL[2,y] := L[y];
    LL[y-2,y] := L[y-2] + 1;
    LL[y-1,y] := L[y-1] + 1;
    DD[1,y] := L[y];
    DD[2,y] := L[y];
    DD[y-2,y] := L[y-2] + 1;
    DD[y-1,y] := L[y-1] + 1;
    for x := 3 to y - 3 do
    begin
      Read(F, c);
      LL[x,y] := L[y] + ord(c) - 48;
      DD[x,y] := LL[x,y];
    end;
    Read(F, c); {space}
    while c = ' ' do Read(F, c);
    if c <> '>' then
    begin
      showmessage('4:Wrong structure at '+inttostr(y)+', #'+ inttostr(ord(c))+' instead of <space>');
      exit;
    end;
    for x := 1 to y do
    begin
      Read(F, c);
      ChainData[x,y] := ord(c) - 64;
    end;
    Readln(F);
  end;
  CloseFile(F);
  TableU.F_Table.M_Table.Lines.add('  n L(n) LL(m,n) - L(n)');
  TableU.F_Table.M_Table.Lines.add('-----------------------');
  for y := 1 to maxLL do
  begin
    s := '  ' + inttostr(y);
    s := copy(s, length(s) - 2, 3);
    t := '  ' + inttostr(ChainU.L[y]);
    t := copy(t, length(t) - 2, 3);
    s := s + '  ' + t + ' ';
    for x_width := 0 to y div CHAINWIDTH do
    begin
      for x := 1 to CHAINWIDTH do
        if x_width * CHAINWIDTH + x < y then
          s := s + inttostr(LL[x_width*CHAINWIDTH+x,y] - L[y]);
      s := s + ' ';
    end;
    TableU.F_Table.M_Table.Lines.add(s);
  end;
  TableU.F_Table.M_Table.Lines.add('-----------------------');
  TableU.F_Table.M_Table.Lines.add('---   Chain Data    ---');
  TableU.F_Table.M_Table.Lines.add('-----------------------');
  TableU.F_Table.M_Table.Lines.add('  n L(n) Chaindata');
  TableU.F_Table.M_Table.Lines.add('-----------------------');
  for y := 1 to maxLL do
  begin
    s := '  ' + inttostr(y);
    s := copy(s, length(s) - 2, 3);
    t := '  ' + inttostr(ChainU.L[y]);
    t := copy(t, length(t) - 2, 3);
    s := s + ' ' + t + '  ';
    for x := 1 to y do
    begin
      t := ' ' + inttostr(ChainData[x,y]);
      t := copy(t, length(t) - 1, 2);
      s := s + t + '.';
    end;
    TableU.F_Table.M_Table.Lines.add(s);
  end;
end;

procedure TF_Chain.B_CalcClick(Sender: TObject);
var
  len, x, y, i : integer;
  s : string;
begin
  cnt := 0;
  status := 1;
  if S_N.value > maxLL then
  begin
    ShowMessage('The table must be made larger!');
    exit;
  end;
  x := S_M.value;
  y := S_N.value;
  if x >= y then exit;
  if (LL[x,y] <= 0) OR (DD[x,y] <= 0) then
  begin
    DebugU.F_Debug.M_Debug.Lines.Add('cannot start Calculation since zero');
    exit;
  end;
  len := calcLL(x, y, DD[x,y], LL[x,y]+1);
  s := '(';
  for i := 0 to len - 1 do
    s := s + inttostr(chain[i]) + ',';
  s := s + inttostr(chain[len]) + ')';
  E_Chain.Text := s;
  E_Len.Text := inttostr(len);
end;

procedure TF_chain.B_TableClick(Sender: TObject);
begin
  F_Table.ShowModal;
end;

procedure TF_Chain.S_Exit(Sender: TObject);
var
  chnge : integer;
begin
  if S_M.value >= S_N.value then
  begin
    chnge := S_M.value;
    S_M.value := S_N.value;
    S_N.value := chnge;
    showmessage('M must be smaller than N.');
  end;
end;

procedure TF_Chain.S_Change(Sender: TObject);
begin
  E_Chain.Text := '()';
  E_Len.Text := '0';
end;

procedure TF_Chain.B_DebugClick(Sender: TObject);
begin
  F_Debug.Show;
end;

procedure TF_Chain.B_ExitClick(Sender: TObject);
begin
  Close;
end;

end.
